/*****************************************************************************/
/*
 *      tiglusb.c -- Texas Instruments' GraphLink USB Driver.
 *      
 *      Copyright (C) 2001 Julien BLACHE <jb@technologeek.org>
 *      Based on the dabusb.c driver, see below.
 *
 *      ========
 *      dabusb.c  --  dab usb driver.
 *
 *      Copyright (C) 1999  Deti Fliegl (deti@fliegl.de)
 *      ========
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *      ======================================================================
 * 
 *      tiglusb.c ChangeLog :
 *      
 *      * 20010927 (v0.01b) : Device probing, configuration, claiming.
 *      * 20010927 (v0.02b) : Bumped MAXTIGL from 1 to 16.
 *      * 20010929 (v0.03b) : - Added a semaphore to tiglusb_t
 *                            - Added the support for this semaphore
 *      * 20010929 (v0.04b) : After probing the device, we use the remote
 *                            control protocol to display the letter 'A'
 *                            on the screen. This was only a test to confirm
 *                            that we only need to send the raw TI protocol
 *                            to the bulk endpoint 2.
 *      * 20010930 (v0.05b) : Added devfs support.
 *      * 20010930 (v0.06b) : First implementation of the *_fops_* functions.
 *                            Not tested yet. Probably buggy :] Well, it actually
 *                            kills the kernel :P
 *      * 20010930 (v0.07b) : Does not kill the kernel anymore when reading the
 *                            answer from the calc at probe (for devel. only).
 *                            The crash was due to the change of the interrupt
 *                            callback (used in the tiglusb_fops_open() function,
 *                            and thus modified), so there's now a
 *                            bulk_tmp_rcv_int_complete() function that is used
 *                            only during the probe (this is for devel only, as
 *                            said earlier).
 *      * 20010930 (v0.08b) : Initialize the interrupt bulk IN before sending
 *                            any data to the calc (already done for the normal
 *                            code via the tiglusb_fops_open() function) in the
 *                            probe function.
 *      * 20011002 (v0.09b) : It works *really* better when the f*cking context
 *                            pointer is actually initialized in the INT_URB
 *                            (in tiglusb_fops_open).
 *      * 20011002 (v0.10b) : Changed down() to down_interruptible() in 
 *                            tiglusb_fops_read().
 *      * 20011002 (v0.10b) : - Do not lock the mutex in tiglusb_fops_read()
 *                            - Added a remove_pending variable to the struct
 *                            - Check for remove_pending in tiglusb_fops_read()
 *                            - Wake up tiglusb_fops_read() in tiglusb_disconnect()
 *                              after setting remove_pending.
 *                            - tiglusb_fops_write() seems to work as expected.
 *                            - tiglusb_fops_read() could work as expected, but not
 *                              in NON_BLOCK mode.
 *                            - Maybe add a mutex preventing tiglusb_disconnect()
 *                              to disconnect the device if some operations are not
 *                              terminated.
 *     * 20011004 (v0.11b) : Rewritten the for() loop in tiglusb_fops_read().
 *                           As a consequence, I can't crash the kernel anymore
 *                           by Ctrl-C'ing a tail -f on /dev/usb/tigl0 and then
 *                           disconnecting the device.
 *                           Unfortunately on disconnect(), the loop continues...
 *                           in other words, there's a bug :P
 *     * 20011004 (v0.12b) : - Some changes to tiglusb_fops_read(), in the way
 *                           we handle 0_NONBLOCK. (namely, changed labels and gotos).
 *                           - Fixed the interrupt callback to actually update s->rcv_pos...
 *                           - Added a check in tiglusb_fops_read() to verify that
 *                           u != s->rcv_pos after wake up. If not, we've probably been
 *                           woken up by the kernel, so abort the read();
 *     * 20011004 (v0.13b) : Do not rely on *ppos in tiglusb_fops_read(), it would become
 *                           meaningless when a read() is aborted.
 *     * 20011006 (v0.14b) : - Do not use FILL_INT_URB, we don't need that, and it
 *                           does not work with our device. It only helps crahsing
 *                           the kernel.
 *                           - Changed FILL_INT_URB to FILL_BULK_URB in tiglusb_fops_open(),
 *                           renamed rcv_int_urb to rcv_in_urb, s->first_rcv_int_urb becomes
 *                           s->rcv_in_urb.
 *     * 20011006 (v0.15b) : - in write() we must use endpoint 2, not 1.
 *                           - in write() we must use usb_sndbulkpipe(), not
 *                           usb_rcvbulkpipe().
 *                           - in disconnect(), use down_interruptible().
 *                           - renamed bulk_rcv_int_complete() (see below).
 *                           - do not lock the mutex in bulk_rcv_complete().
 *                           - removed unused snd_buf and snd_pos from s.
 *     * 20011007 (v0.16b) : - added s->read_rcv_pos, do not rely on *ppos at
 *                           all in tiglusb_fops_read(). Pisses me off.
 *                           - fixed the case where u == rcv_pos in tiglusb_fops_read()
 *
 */

/*****************************************************************************/

/* WE SHOULD START AND STOP THE BULK INTERRUPT IN IN READ() AND NOT IN OPEN () */


/*
 * Compile with : (on PPC)
gcc -D__KERNEL__ -I/usr/src/linux-2.4.9-ppc-xfs/include  -Wall -Wstrict-prototypes -Wno-trigraphs -O2 -fno-strict-aliasing -fno-common -fomit-frame-pointer -D__powerpc__ -fsigned-char -msoft-float -pipe -ffixed-r2 -Wno-uninitialized -mmultiple -mstring -DMODULE   -c -o tiglusb.o tiglusb.c
 */

#include <linux/module.h>
#include <linux/socket.h>
#include <linux/miscdevice.h>
#include <linux/list.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>
#include <linux/delay.h>
#include <linux/usb.h>
#include <linux/smp_lock.h>
#include <linux/devfs_fs_kernel.h>

#include "tiglusb.h"


/* --------------------------------------------------------------------- */

/*
 * urb_submit callback on endpoint 2
 */

void tiglusb_bulk_snd_complete (struct urb *purb)
{
  printk("TIGL: tiglusb_bulk_snd_complete\n");

  kfree(purb->transfer_buffer);

  usb_free_urb(purb);

  printk("TIGL: tiglusb_bulk_snd_complete: URB freed\n");
}

/*
 * Interrupt callback
 * -> woke up the tiglusb_fops_read
 */


#if 0
/*
 * THIS IS A TMP FUNCTION INTENDED FOR DEVELOPMENT ONLY
 */

void tiglusb_tmp_bulk_rcv_int_complete (struct urb *purb)
{
  char *buf = purb->transfer_buffer;

  printk("TIGL: TEMP bulk_rcv_int_complete\n");

  printk("answer : 0x%x 0x%x 0x%X 0x%x\n", buf[0], buf[1], buf[2], buf[3]);

  kfree(buf);
  usb_free_urb(purb);
}
#endif 0


void tiglusb_bulk_rcv_complete (struct urb *purb)
{
  char *buf;
  ptiglusb_t s;
  int i;

  printk("TIGL: INT_CB\n");

  printk("TIGL: INT_CB: getting *buffer\n");
  
  buf = (char *) purb->transfer_buffer;

  printk("TIGL: INT_CB: getting *context\n");

  s = (ptiglusb_t) purb->context;

  /* debug */
  if (buf == NULL)
    printk("TIGL: INT_CB: AIIIEEEEEE *buf is NULL !!\n");

  if (s == NULL)
    printk("TIGL: INT_CB: AIIIIEEEEE *s is NULL !!\n");
  /* */

  printk("TIGL: INT_CB: LOCKING MUTEX\n");

  // down(&s->mutex); /* using down() here will result in the kernel killing the interrupt handler */

  printk("TIGL: INT_CB: LOCKED MUTEX\n");

  /*
   * Copy the data to our (2 * BULK_RCV_MAX) cyclic buffer
   */

  printk("TIGL: INT_CB: copying buffer\n");

  printk("TIGL: INT_CB: bytes received: ");

  for (i = 0; i < purb->actual_length; i++)
    {
      /* initiate s->rcv_pos at 1 */
      s->rcv_pos = (s->rcv_pos + 1) % (2 * BULK_RCV_MAX);
      
      printk("0x%x (i = %d, s->rcv_pos = %d) ", buf[i], i, s->rcv_pos);

      s->rcv_buf[s->rcv_pos] = buf[i];
    }

  printk("\n");
      
  printk("TIGL: INT_CB: RELEASING MUTEX\n");

  //up(&s->mutex); /* release */

  printk("TIGL: INT_CB: waking up tiglusb_fops_read\n");

  wake_up(&s->wait); /* wake up the tiglusb_fops_read() */

  /*
   * Reset some of the URB's datas
   */

  printk("TIGL: INT_CB: resetting URB (memset)\n");

  memset(buf, 0, BULK_RCV_MAX);

  printk("TIGL: INT_CB: resetting URB (actual_length)\n");

  purb->actual_length = 0;

  printk("TIGL: INT_CB: re-submitting URB\n");

  usb_submit_urb(purb);
  
  printk("TIGL: INT_CB: URB re-submitted\n");
}

/* --------------------------------------------------------------------- */

/*
 * file operations-related functions
 */

static ssize_t tiglusb_fops_read(struct file *file, char *ubuf, size_t count, loff_t * ppos)
{
  ptiglusb_t s = (ptiglusb_t) file->private_data;

  char *kbuf;
  
  int i;
  int u;
  int ret;

  int back_pos;

  printk("TIGL: tiglusb_fops_read\n");

  if (s->remove_pending)
    {
      printk("TIGL: tiglusb_fops_read: remove pending\n");
      return -EIO;
    }

  if (!s->usbdev)
    return -EIO;

  kbuf = (char *) kmalloc(count, GFP_KERNEL);
  
  if (!kbuf)
    {
      return -ENOMEM;
    }
  
  memset(kbuf, 0, count);

  /*
   * the following code may _REALLY_ be *BUGGY*. You've been warned.
   */

  printk("TIGL: preparing buffer, application requested %d bytes...\n", count);

  back_pos = s->read_rcv_pos;

  for (i = 0; i < count; i++)
    {
      s->read_rcv_pos ++;

      u = s->read_rcv_pos % (2 * BULK_RCV_MAX);

      printk("TIGL: tiglusb_fops_read: u = %d, s->read_rcv_pos = %d\n", u, s->read_rcv_pos);
      
      if (u == s->rcv_pos) /* we're at end of buffer */
	{
	  printk("TIGL: tiglusb_fops_read: u == rcv_pos\n");

	  if (file->f_flags & O_NONBLOCK)  /* opened O_NONBLOCK */
	    {
	      printk("TIGL: tiglusb_fops_read: read() O_NONBLOCK\n");
	      
	      if (i == 0)
		ret = -EAGAIN;  /* no data to transfer, try again */
	      else
		ret = i + 1;   /* i + 1 bytes transferred */

	      printk("TIGL: tiglusb_fops_read: moving 0x%x to kbuf (in u == rcv_pos)\n", s->rcv_buf[u]);
	      kbuf[i] = s->rcv_buf[u];

	      goto out;
	    }
	  else
	    {
	      printk("TIGL: tiglusb_fops_read: sleeping...\n");
	      interruptible_sleep_on(&s->wait);  /* wait new data, blocking mode */
	    }
	}

      if (u == s->rcv_pos)  /* seems we got woken up by the kernel -- happens when Ctrl-C'ing an ongoing read() */
	{
	  printk("TIGL: tiglusb_fops_read: got woken up by kernel, read() aborted...\n");

	  ret = -EIO;
	  goto err;
	}

      if (s->remove_pending)  /* device is to be removed */
	{
	  printk("TIGL: tiglusb_fops_read: remove pending...\n");
	  
	  ret = -EIO;
	  goto err;
	}

      printk("TIGL: tiglusb_fops_read: moving 0x%x to kbuf\n", s->rcv_buf[u]);
      kbuf[i] = s->rcv_buf[u];
    }

  ret = count;

 out:
  if (i == -EAGAIN)
    goto err;

  printk("TIGL: copying buffer to user-space...\n");
  
  if (copy_to_user(ubuf, kbuf, ret))
    {
      printk("TIGL: tiglusb_fops_read: error in copy_to_user\n");

      ret = -EFAULT;
      goto err;
    }

 err:
  printk("TIGL: tiglusb_fops_read: end of function\n");
  kfree(kbuf);
  *ppos = *ppos + (s->read_rcv_pos - back_pos);
  return ret;
}

static ssize_t tiglusb_fops_write(struct file *file, const char *ubuf, size_t count, loff_t * ppos)
{
  ptiglusb_t s = (ptiglusb_t) file->private_data;

  char *kbuf;

  int ret;

  purb_t snd_bulk_urb;

  printk("TIGL: tiglusb_fops_write\n");
  
  if (!s->usbdev)
    return -EIO;
  
  printk("TIGL: tiglusb_fops_write: LOCKING MUTEX\n");

  down(&s->mutex); /* lock */
  
  printk("TIGL: tiglusb_fops_write: MUTEX LOCKED, allocating memory\n");

  kbuf = (char *) kmalloc(count, GFP_KERNEL);

  if (!kbuf)
    {
      up(&s->mutex);
      return -ENOMEM;
    }

  memset(kbuf, 0, count);

  printk("TIGL: copying buffer to kernel-space...\n");

  /*
   * Copy the data
   */
  
  if (copy_from_user(kbuf, ubuf, count))
    {
      printk("TIGL: tiglusb_fops_write: error in copy_from_user\n");
      
      ret = -EFAULT;
      goto out;
    }

  /*
   * Fill the URB
   */
  
  printk("TIGL: tiglusb_fops_write: writing 0x%x\n", kbuf[0]);

  printk("TIGL: tiglusb_fops_write: filling URB\n");

  snd_bulk_urb = usb_alloc_urb(0);
  
  FILL_BULK_URB(snd_bulk_urb,
		s->usbdev,
		usb_sndbulkpipe(s->usbdev, 2),
		kbuf,
		count,
		tiglusb_bulk_snd_complete,
		NULL); /* do we need a context ? */

  /*
   * There's no timeout set, maybe we'll need one
   */

  printk("TIGL: tiglusb_fops_write: submitting URB\n");

  ret = usb_submit_urb(snd_bulk_urb); /* we must take care of this return value */

  printk("TIGL: tiglusb_fops_write: URB submitted\n");


  /*
   * Taken from dabusb.c
   */

  if (ret < 0)
    {
      if (ret == -EPIPE)
	{
	  printk("TIGL: tiglusb_fops_write: trying to clear STALL condition\n");
	  
	  if (usb_clear_halt(s->usbdev, usb_pipeendpoint(snd_bulk_urb->pipe)))
	    printk("TIGL: tiglusb_fops_write: could not clear STALL condition\n");
	}
      else
	printk("TIGL: tiglusb_fops_write: ERROR: %d\n", ret);
    }
  else
    {
      ret = count;
      
      *ppos = *ppos + count;
    }

  printk("TIGL: tiglusb_fops_write: reached end of function\n");

 out:
  up(&s->mutex); /* release */
  return ret;
}

static int tiglusb_fops_open(struct inode *inode, struct file *file)
{
  int num = MINOR (inode->i_rdev);
  ptiglusb_t s;

  char *urb_buf;
  purb_t rcv_in_urb;
  
  printk("TIGL: tiglusb_fops_open, minor = %d\n", num);

  if (num < TIGLUSB_MINOR || num >= (TIGLUSB_MINOR + MAXTIGL))
    return -EIO;
  
  s = &tiglusb[num - TIGLUSB_MINOR];

  down (&s->mutex); /* lock */

  /*
   * No device or device already opened
   */

  while (!s->usbdev || s->opened)
    {
      up (&s->mutex);
      
      if (file->f_flags & O_NONBLOCK)
	{
	  return -EBUSY;
	}

      down (&s->mutex);
    }


  /*
   * Mark the device as opened
   */

  printk("TIGL: tiglusb_fops_open: loop exited\n");

  s->read_rcv_pos = 0;

  s->opened = 1;

  /*
   * Initiate the interrupt-driven reception
   */
  
  printk("TILP: tiglusb_fops_open: initiating BULK IN...\n");
  
  urb_buf = (char *)kmalloc(BULK_RCV_MAX, GFP_KERNEL);
  
  memset (urb_buf, 0, BULK_RCV_MAX);
  
  rcv_in_urb = usb_alloc_urb(0); /* maybe we should link several URBs... */
  
  FILL_BULK_URB(rcv_in_urb,
		s->usbdev,
		usb_rcvbulkpipe(s->usbdev, 1),
		urb_buf,
		BULK_RCV_MAX,
		tiglusb_bulk_rcv_complete,
		s);

  usb_submit_urb(rcv_in_urb); /* maybe we should take care of this return code */
  
  printk("TIGL: tiglusb_fops_open: URB submitted\n");

  s->rcv_in_urb = rcv_in_urb; /* keep track of the first URB */

  up (&s->mutex);
  
  file->f_pos = 0;
  file->private_data = s;
  
  printk("TIGL: tiglusb_fops_read: reached end of function\n");

  return 0;
}

static int tiglusb_fops_release(struct inode *inode, struct file *file)
{
  ptiglusb_t s = NULL;
  purb_t rcv_in_urb;

  printk("TIGL: tiglusb_fops_release\n");

  /*
   * Free the URBs
   */

  if (file->private_data != NULL)
    {
      s = (ptiglusb_t) file->private_data;

      printk("TIGL: tiglusb_fops_release: LOCKING MUTEX\n");

      down(&s->mutex);

      printk("TIGL: tiglusb_fops_release: MUTEX LOCKED\n");

      rcv_in_urb = s->rcv_in_urb;
      s->rcv_in_urb = NULL;
      
      printk("TIGL: tiglusb_fops_release: freeing transfer buffer\n");

      kfree(rcv_in_urb->transfer_buffer);
      
      printk("TIGL: tiglusb_fops_release: freeing URB\n");

      usb_free_urb(rcv_in_urb);

      printk("TIGL: tiglusb_fops_release: URB freed\n");
      
      s->read_rcv_pos = 0;

      /*
       * Mark the device as closed
       */
      
      s->opened = 0;

      up(&s->mutex);
    }
  else
    printk("TIGL: tiglusb_fops_release: file->private_data was NULL\n");

  return 0;
}

static struct file_operations tiglusb_fops =
{
	owner:		THIS_MODULE,
	read:		tiglusb_fops_read,
	write:          tiglusb_fops_write,
	open:           tiglusb_fops_open,
	release:	tiglusb_fops_release,
};

/* --------------------------------------------------------------------- */
static void *tiglusb_probe (struct usb_device *usbdev, unsigned int ifnum,
			   const struct usb_device_id *id)
{
  ptiglusb_t s;
  unsigned int u;

  char nodename[7];

  /*****************/
#if 0

  int err;
  char *test_string = NULL;

  purb_t test_urb;

  char *rcv_string = NULL;
#endif
  /*****************/

  printk("TIGL: PROBING DEVICE\n");
  printk("TIGL: tiglusb_probe: vendor id 0x%x, device id 0x%x ifnum:%d\n",
	 usbdev->descriptor.idVendor, usbdev->descriptor.idProduct, ifnum);
  
  /* 
   * We don't handle multiple configurations
   * As of version 0x0103 of the TIGL hardware, there's only 1 configuration 
   */
  
  if (usbdev->descriptor.bNumConfigurations != 1)
    {
      printk("TIGL: tiglusb_probe: More than 1 configurations, giving up\n");
      return NULL;
    }

  if ((usbdev->descriptor.idProduct == 0xe001) && (usbdev->descriptor.idVendor == 0x451)) 
    {
      printk("TIGL: tiglusb_probe: found TIGL USB pid 0x%x\n", usbdev->descriptor.idVendor);
      
      printk("TIGL: tiglusb_probe: attempting to set configuration\n");
      
      if (usb_set_configuration (usbdev, usbdev->config[0].bConfigurationValue) < 0)
	{
	  printk("TIGL: tiglusb_probe: set_configuration failed\n");
	  goto reject;
	}
      
      printk("TIGL: tiglusb_probe: set_configuration successful\n");
      
      /*
       * find a free struct
       */
      
      for (u = 0; u < MAXTIGL; u++)
	{
	  s = &tiglusb[u];
	  if (!s->usbdev)
	    break;
	  else
	    s = NULL;
	}
      
      if (s == NULL)  /* we already support 16 devices */
	goto reject;
      
      down(&s->mutex); /* locking */
      
      s->usbdev = usbdev;
    }
  else
    {
      goto reject;
    }
  
  printk("TIGL: tiglusb_probe: bound to interface: %d\n", ifnum);

  /***************/

  printk("TIGL: *** establishing comm ... ***\n");
 
#if 0
 
#define TAILLE_DE_RCV_STRING 32
  
  printk("Reading calc answer...\n");
  
  rcv_string = (char *) kmalloc(TAILLE_DE_RCV_STRING, GFP_KERNEL);
  
  memset (rcv_string, 0, TAILLE_DE_RCV_STRING);
  
  test_urb = usb_alloc_urb(0);
  
  FILL_BULK_URB(test_urb,
	       s->usbdev,
	       usb_rcvbulkpipe(usbdev, 1),
	       rcv_string,
	       TAILLE_DE_RCV_STRING,
	       tiglusb_tmp_bulk_rcv_int_complete, /* this function will be removed upon release */
		NULL);
  
  err = usb_submit_urb(test_urb);


  /* maybe we should take care of this return code */
  
#define TAILLE_DE_TEST_STRING 4
  
  printk("Sending remote control to calc (display \"A\")...\n");
  
  test_string = (char *)kmalloc(TAILLE_DE_TEST_STRING,
				GFP_KERNEL);
  
  memset (test_string, TAILLE_DE_TEST_STRING, 0);
  
  test_string[0] = 0x08;
  test_string[1] = 0x87;
  test_string[2] = 'A';
  test_string[3] = 0;
  
  printk("TS: %s\n", test_string);
  
  err = usb_bulk_msg(usbdev,
		     usb_sndbulkpipe(usbdev, 2),
		     test_string,
		     TAILLE_DE_TEST_STRING,
		     NULL,
		     64444);
  
  kfree(test_string);

  if (err)
    {
      printk("Encountered error... %d\n", err);
    }
  
  printk("Sent.\n");

#endif
  
  /****************/

  sprintf(nodename, "tigl%d", s->devnum);
  
  printk("Registering to devfs : major = %d, minor = %d, nodename = %s\n", USB_MAJOR, (TIGLUSB_MINOR + s->devnum), nodename);

  s->devfs = devfs_register(usb_devfs_handle, nodename,
			    DEVFS_FL_DEFAULT, USB_MAJOR,
			    TIGLUSB_MINOR + s->devnum,
			    S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP |
			    S_IWGRP, &tiglusb_fops, NULL);
  
  printk("TIGL: New TIGL USB registered, device %d\n", s->devnum);

  /*
   * Initialize rcv_buf (arbitrarily 2 * BULK_*_MAX)
   */

  s->remove_pending = 0;
  
  s->rcv_buf = (char *)kmalloc(2 * BULK_RCV_MAX, GFP_KERNEL);
  memset(s->rcv_buf, 0, 2 * BULK_RCV_MAX);
  s->rcv_pos = 0;
  
  up(&s->mutex); /* release locking */
  
  MOD_INC_USE_COUNT;
  
  /*
   * The device is probed, configured, we're now indicating
   * to the USB susbsystem that this driver handles this device
   * from now on.
   */
  
  return s;
  
 reject:
  printk("TIGL: tiglusb_probe: REJECTED !\n");
  return NULL;
}

static void tiglusb_disconnect (struct usb_device *usbdev, void *drv_context)
{
  ptiglusb_t s = NULL;

  dbg("TIGL: tiglusb_disconnect");

  s = (ptiglusb_t) drv_context;
  
  down_interruptible(&s->mutex); /* lock */

  printk("TIGL: tiglusb_disconnect: remove_pending = 1\n");

  s->remove_pending = 1;

  printk("TIGL: tiglusb_disconnect: waking up sleeping functions\n");
  
  wake_up(&s->wait);

  s->usbdev = NULL;

  s->opened = 0;
  
  /*
   * Unregister the devfs stuff
   */

  printk("TIGL: unregistering to devfs\n");

  devfs_unregister(s->devfs);

  s->devfs = NULL;

  /*
   * Free the buffers
   */

  printk("TIGL: tiglusb_disconnect: freeing rcv_buf\n");

  s->rcv_pos = 0;
  kfree(s->rcv_buf);
  
  printk("TIGL: tiglusb_disconnect: buffers freed\n");

  up(&s->mutex); /* release */

  /*
   * Verify that by decreasing USE_COUNT it won't become < 0
   */

  if (GET_USE_COUNT(THIS_MODULE) > 0)
    {
      MOD_DEC_USE_COUNT;
    }

  printk("TIGL: device disconnected\n");
}

static struct usb_device_id tiglusb_ids [] = {
	{ USB_DEVICE(0x0451, 0xe001) },
	{ }						/* Terminating entry */
};

MODULE_DEVICE_TABLE (usb, tiglusb_ids);

static struct usb_driver tiglusb_driver =
{
  name:		"tiglusb.c",
  probe:		tiglusb_probe,
  disconnect:	tiglusb_disconnect,
  fops:		&tiglusb_fops,
  minor:		TIGLUSB_MINOR,
  id_table:	tiglusb_ids,
};

/* --------------------------------------------------------------------- */

static int __init tiglusb_init (void)
{
  unsigned u;
  
  /* initialize struct */
  for (u = 0; u < MAXTIGL; u++)
    {
      ptiglusb_t s = &tiglusb[u];
      memset (s, 0, sizeof (tiglusb_t));
      init_MUTEX (&s->mutex);
      s->usbdev = NULL;
      s->devnum = u;
      s->opened = 0;
      init_waitqueue_head (&s->wait);
      //init_waitqueue_head (&s->remove_ok);
      //spin_lock_init (&s->lock);
    }

  /*
   * Registering to the USB subsystem
   */
  
  if (usb_register(&tiglusb_driver))
    return -1;
  
  dbg("TIGL: tiglusb_init: driver registered");
  
  info(DRIVER_VERSION ":" DRIVER_DESC);
  
  return 0;
}

static void __exit tiglusb_cleanup (void)
{
  dbg("TIGL: tiglusb_cleanup: cleaning up");
  
  usb_deregister (&tiglusb_driver);
  
  dbg("TIGL: tiglusb_cleanup: cleanup terminated");
}

/* --------------------------------------------------------------------- */

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);

module_init (tiglusb_init);
module_exit (tiglusb_cleanup);

/* --------------------------------------------------------------------- */
